from __future__ import annotations

import contextlib

from argrelay.enum_desc.PluginType import PluginType
from argrelay.plugin_interp.AbstractInterpFactory import AbstractInterpFactory
from argrelay.plugin_interp.InterpTreeInterp import InterpTreeInterp
from argrelay.plugin_interp.InterpTreeInterpFactoryConfigSchema import (
    tree_path_interp_factory_config_desc,
    interp_selector_tree_,
)
from argrelay.plugin_interp.TreeWalker import TreeWalker
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_data.ServerConfig import ServerConfig, assert_plugin_instance_id


class InterpTreeInterpFactory(AbstractInterpFactory):
    """
    Implements FS_01_89_09_24 interp tree.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )
        self.is_recursive_func_load = False

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        return tree_path_interp_factory_config_desc.dict_from_input_dict(plugin_config_dict)

    def load_func_envelopes(
        self,
        interp_tree_abs_path: tuple[str, ...],
        func_ids_to_func_envelopes: dict[str, dict],
    ) -> list[str]:
        with self._is_recursive_load() as is_recursive_load:
            if is_recursive_load:
                return []
            return self._load_func_envelopes(
                interp_tree_abs_path,
                func_ids_to_func_envelopes,
            )

    @contextlib.contextmanager
    def _is_recursive_load(self):
        """
        Auto-reset flag `self.is_recursive_func_load` on exit from `with` section.
        """
        try:
            if not self.is_recursive_func_load:
                self.is_recursive_func_load = True
                is_recursive_load = False
            else:
                is_recursive_load = True
            yield is_recursive_load
        finally:
            self.is_recursive_func_load = False

    def _load_func_envelopes(
        self,
        interp_tree_abs_path: tuple[str, ...],
        func_ids_to_func_envelopes: dict[str, dict],
    ) -> list[str]:
        mapped_func_ids: list[str] = super().load_func_envelopes(
            interp_tree_abs_path,
            func_ids_to_func_envelopes,
        )
        tree_walker: TreeWalker = TreeWalker(
            "interp_tree",
            self.plugin_config_dict[interp_selector_tree_],
        )
        # Walk configured interp tree and call `load_func_envelopes` with `interp_tree_abs_path` for each interp.
        interp_rel_paths: dict[str, list[list[str]]] = tree_walker.build_str_leaves_paths()
        for interp_plugin_id in interp_rel_paths:
            for interp_rel_path in interp_rel_paths[interp_plugin_id]:
                assert_plugin_instance_id(
                    self.server_config,
                    interp_plugin_id,
                    PluginType.InterpFactoryPlugin,
                )
                interp_factory: AbstractInterpFactory = self.server_config.interp_factories[interp_plugin_id]
                sub_interp_tree_abs_path = interp_tree_abs_path + tuple(interp_rel_path)

                mapped_func_ids.extend(interp_factory.load_func_envelopes(
                    sub_interp_tree_abs_path,
                    func_ids_to_func_envelopes,
                ))
        return mapped_func_ids

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> InterpTreeInterp:
        """
        Each `InterpTreeInterpFactory` serves a subtree,
        but it only stores config of per its plug points
        (interp tree abs paths where it was invoked on `load_func_envelopes`).
        If next interp with its plugin id is created, but the specified path is within its subtree,
        chop off the last step in the path until it matches path to its subtree to select one.
        For example:
        ```
        # Current path within interp tree:
        interp_ctx.interp_tree_abs_path = ('l1', 'l2', 'l3', 'l4', )
        # Plug point of this factory within the interp tree:
        self.interp_tree_abs_paths_to_node_configs = {
            ('l1', 'l2', ): {},
            ('x1', 'x2', 'x3', ): {},
        }
        # Keep trying to cut of the last step in the path until it matches one of the key:
        ('l1', 'l2', 'l3', 'l4', )
        ('l1', 'l2', 'l3', )
        ('l1', 'l2', )
        # Done.
        ```
        """
        orig_subtree_abs_path = interp_ctx.interp_tree_abs_path
        curr_subtree_abs_path = orig_subtree_abs_path
        while curr_subtree_abs_path not in self.interp_tree_abs_paths_to_node_configs:
            if len(curr_subtree_abs_path) == 0:
                raise ValueError(
                    f"path `{orig_subtree_abs_path}` is not subpath of `{self.interp_tree_abs_paths_to_node_configs.keys()}` paths"
                )
            curr_subtree_abs_path = curr_subtree_abs_path[:-1]
        return InterpTreeInterp(
            self.plugin_instance_id,
            self.interp_tree_abs_paths_to_node_configs[curr_subtree_abs_path],
            interp_ctx,
        )



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.enum_desc.DistinctValuesQuery import DistinctValuesQuery
from argrelay.mongo_data.MongoClientConfig import MongoClientConfig
from argrelay.mongo_data.MongoServerConfig import MongoServerConfig


@dataclass
class MongoConfig:
    use_mongomock: bool = field()
    """
    See :class:`MongoConfigSchema.use_mongomock`.
    """

    distinct_values_query: DistinctValuesQuery = field()
    """
    See `DistinctValuesQuery`.
    """

    mongo_client: MongoClientConfig = field()

    mongo_server: MongoServerConfig = field()



"""
"""

goto_host_func_ = "goto_host_func"
goto_service_func_ = "goto_service_func"
goto_repo_func_ = "goto_repo_func"

desc_host_func_ = "desc_host_func"
desc_service_func_ = "desc_service_func"

list_host_func_ = "list_host_func"
list_service_func_ = "list_service_func"

desc_commit_func_ = "desc_commit_func"



from __future__ import annotations

from marshmallow import Schema, fields, RAISE, post_load

from argrelay.custom_integ.ServiceArgType import ServiceArgType
from argrelay.custom_integ.ServiceEnvelopeClass import ServiceEnvelopeClass
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_data.EnvelopeCollection import EnvelopeCollection
from argrelay.runtime_data.PluginEntry import PluginEntry
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_core_client.ConnectionConfigSchema import connection_config_desc
from argrelay.schema_config_core_server.GuiBannerConfigSchema import gui_banner_config_desc
from argrelay.schema_config_core_server.MongoConfigSchema import mongo_config_desc
from argrelay.schema_config_core_server.QueryCacheConfigSchema import query_cache_config_desc
from argrelay.schema_config_core_server.ServerPluginControlSchema import server_plugin_control_desc
from argrelay.schema_config_core_server.StaticDataSchema import static_data_desc
from argrelay.schema_config_plugin.PluginEntrySchema import plugin_entry_desc

connection_config_ = "connection_config"
mongo_config_ = "mongo_config"
query_cache_config_ = "query_cache_config"
gui_banner_config_ = "gui_banner_config"
class_to_collection_map_ = "class_to_collection_map"
server_plugin_control_ = "server_plugin_control"
plugin_instance_entries_ = "plugin_instance_entries"
static_data_ = "static_data"


class ServerConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    connection_config = fields.Nested(
        connection_config_desc.dict_schema,
        required = True,
    )

    mongo_config = fields.Nested(
        mongo_config_desc.dict_schema,
        required = True,
    )

    query_cache_config = fields.Nested(
        query_cache_config_desc.dict_schema,
        required = True,
    )

    gui_banner_config = fields.Nested(
        gui_banner_config_desc.dict_schema,
        required = True,
    )

    # Related to FS_56_43_05_79 search diff collection:
    # Maps names of envelope classes to names of `EnvelopeCollection`.
    # Specifically, it allows to:
    # *   map classes to put all into single collection
    # *   map classes to put each into separate collection
    # Normally, if map entry is missing for one of the envelope class,
    # it uses envelope class name as collection name by default.
    class_to_collection_map = fields.Dict(
        keys = fields.String(),
        values = fields.String(),
        required = True,
    )

    server_plugin_control = fields.Nested(
        server_plugin_control_desc.dict_schema,
        required = True,
    )

    # Plugin config data: key = `plugin_instance_id`, value = `plugin_entry`:
    plugin_instance_entries = fields.Dict(
        keys = fields.String(),
        values = fields.Nested(plugin_entry_desc.dict_schema),
        required = True,
    )

    # TODO_00_79_72_55: remove in the future:
    static_data = fields.Nested(
        static_data_desc.dict_schema,
        required = False,
        load_default = StaticData(
            envelope_collections = {},
        ),
    )

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        plugin_instance_id_activate_order_dag = {}
        for plugin_instance_id in input_dict[plugin_instance_entries_]:
            plugin_entry: PluginEntry = input_dict[plugin_instance_entries_][plugin_instance_id]
            plugin_instance_id_activate_order_dag[plugin_instance_id] = plugin_entry.plugin_dependencies

        # Populate `plugin_instance_id` from `plugin_instance_entries` into each
        return ServerConfig(
            connection_config = input_dict[connection_config_],
            mongo_config = input_dict[mongo_config_],
            query_cache_config = input_dict[query_cache_config_],
            gui_banner_config = input_dict[gui_banner_config_],
            class_to_collection_map = input_dict[class_to_collection_map_],
            server_plugin_control = input_dict[server_plugin_control_],
            plugin_instance_id_activate_list = serialize_dag_to_list(
                plugin_instance_id_activate_order_dag,
            ),
            plugin_instance_entries = input_dict[plugin_instance_entries_],
            static_data = input_dict[static_data_],
        )


server_config_desc = TypeDesc(
    dict_schema = ServerConfigSchema(),
    ref_name = ServerConfigSchema.__name__,
    dict_example = {
        connection_config_: connection_config_desc.dict_example,
        mongo_config_: mongo_config_desc.dict_example,
        query_cache_config_: query_cache_config_desc.dict_example,
        gui_banner_config_: gui_banner_config_desc.dict_example,
        class_to_collection_map_: {
            ServiceEnvelopeClass.ClassCluster.name: ServiceEnvelopeClass.ClassCluster.name,
            ServiceEnvelopeClass.ClassHost.name: ServiceEnvelopeClass.ClassHost.name,
            ServiceEnvelopeClass.ClassService.name: ServiceEnvelopeClass.ClassService.name,
            ServiceArgType.AccessType.name: ServiceArgType.AccessType.name,
        },
        server_plugin_control_: server_plugin_control_desc.dict_example,
        plugin_instance_entries_: {
            "some_plugin_instance_id": plugin_entry_desc.dict_example,
            "some_plugin_instance_id1": plugin_entry_desc.dict_example,
            "some_plugin_instance_id2": plugin_entry_desc.dict_example,
        },
        static_data_: static_data_desc.dict_example,
    },
    default_file_path = "argrelay.server.yaml",
)


def serialize_dag_to_list(
    entire_dag: dict[str, list[str]],
) -> list[str]:
    output_list: list[str] = list_sub_dag(
        [],
        entire_dag,
        [node_id for node_id in entire_dag],
    )
    return output_list


def list_sub_dag(
    curr_path: list[str],
    entire_dag: dict[str, list[str]],
    id_sub_list: list[str],
) -> list[str]:
    output_list: list[str] = []

    for node_id in id_sub_list:
        if node_id in curr_path:
            raise ValueError(f"cyclic ref to plugin id in path `{curr_path}` -> `{node_id}`")
        if node_id not in entire_dag:
            raise ValueError(f"plugin id in path `{curr_path}` -> `{node_id}` is not defined")
        if node_id in entire_dag:
            # plugin has dependencies:
            sub_list = list_sub_dag(
                curr_path + [node_id],
                entire_dag,
                entire_dag[node_id],
            )
            for sub_node_id in sub_list:
                if sub_node_id not in output_list:
                    output_list.append(sub_node_id)
        else:
            raise ValueError(f"plugin id in path `{curr_path}` -> `{node_id}` is not included for activation")

        if node_id not in output_list:
            output_list.append(node_id)

    return output_list



from argrelay.enum_desc.InterpStep import InterpStep
from argrelay.enum_desc.TokenType import get_token_type, TokenType

from argrelay.runtime_context.InterpContext import InterpContext

# `AbstractInterp` is NOT a plugin `AbstractInterpFactory` is:
class AbstractInterp:
    """
    Interpreter processes command line sharing current state via :class:`InterpContext`.

    New instance of interpreter is created by (a plugin implementing) `AbstractInterpFactory` for each request.
    """

    instance_counter: int = 0
    """
    Instance counter per interp class.
    """

    def __init__(
        self,
        interp_factory_id: str,
        interp_tree_node_config_dict: dict,
        interp_ctx: InterpContext,
    ):
        self.interp_factory_id: str = interp_factory_id

        self.__class__.instance_counter += 1
        self.instance_number: int = self.instance_counter

        self.interp_tree_node_config_dict: dict = interp_tree_node_config_dict
        """
        The configs for individual `AbstractInterp` are called `interp_tree_node_config_dict` and cloned/populated by
        `load_func_envelopes` based on their position in FS_01_89_09_24 interp tree.
        """

        self.interp_ctx: InterpContext = interp_ctx
        self.base_container_ipos: int = interp_ctx.curr_container_ipos

    def __str__(self) -> str:
        return f"fid: {self.interp_factory_id} path: {self.interp_ctx.interp_tree_abs_path} instance: {self.instance_number}"

    def consume_key_args(self) -> None:
        pass

    def consume_pos_args(self) -> None:
        pass

    def try_iterate(self) -> InterpStep:
        pass

    def has_fill_control(
        self,
    ) -> bool:
        """
        FS_72_53_55_13: See `AbstractDelegator.has_fill_control` for more information.
        """
        return False

    def delegate_fill_control(
        self,
    ) -> bool:
        """
        FS_72_53_55_13: See `AbstractDelegator.run_fill_control` for more information.
        """
        return False

    def propose_arg_completion(self) -> None:
        pass

    def next_interp(self) -> "AbstractInterp":
        """
        Return next interp factory id (or None).
        """
        return None

    def is_pos_arg(
        self,
        token_ipos: int,
    ) -> bool:
        return get_token_type(
            self.interp_ctx.parsed_ctx.all_tokens,
            token_ipos,
        ) is TokenType.PosArg



from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.schema_request.CallContextSchema import server_action_
from argrelay.test_infra.BaseTestClass import BaseTestClass
from argrelay.test_infra.JsonTestOutputVerifier import JsonTestOutputVerifier


class CustomVerifier(JsonTestOutputVerifier):

    def __init__(
        self,
        test_instance: BaseTestClass,
    ):
        super().__init__(
        )

        self.test_instance: BaseTestClass = test_instance


class CallContextVerifier(CustomVerifier):

    def __init__(
        self,
        test_instance: BaseTestClass,
        server_action: ServerAction,
    ):
        super().__init__(
            test_instance,
        )

        self.server_action: ServerAction = server_action

        self.add_verifier(
            f"$.{server_action_}",
            lambda m: test_instance.assertEqual(1, len(m)),
            lambda m: test_instance.assertEqual(server_action.name, m[0].value),
        )


class ServerActionVerifier(CallContextVerifier):

    def __init__(
        self,
        test_instance: BaseTestClass,
        server_action: ServerAction,
    ):
        super().__init__(
            test_instance,
            server_action,
        )

        self.test_instance = test_instance,
        self.server_action = server_action,


class ProposeArgValuesVerifier(ServerActionVerifier):

    def __init__(
        self,
        test_instance: BaseTestClass,
    ):
        super().__init__(
            test_instance,
            ServerAction.ProposeArgValues,
        )


class DescribeLineArgsVerifier(ServerActionVerifier):

    def __init__(
        self,
        test_instance: BaseTestClass,
    ):
        super().__init__(
            test_instance,
            ServerAction.DescribeLineArgs,
        )


class RelayLineArgsVerifier(ServerActionVerifier):

    def __init__(
        self,
        test_instance: BaseTestClass,
    ):
        super().__init__(
            test_instance,
            ServerAction.RelayLineArgs,
        )



from __future__ import annotations

import json

from argrelay.custom_integ.ServiceArgType import ServiceArgType
from argrelay.custom_integ.ServiceEnvelopeClass import ServiceEnvelopeClass
from argrelay.custom_integ.value_constants import (
    goto_host_func_,
    goto_service_func_,
    list_host_func_,
    list_service_func_,
    desc_host_func_,
    desc_service_func_,
)
from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.plugin_delegator.AbstractDelegator import (
    AbstractDelegator,
    get_func_id_from_interp_ctx,
    get_func_id_from_invocation_input,
)
from argrelay.plugin_delegator.ErrorDelegator import ErrorDelegator
from argrelay.plugin_delegator.ErrorDelegatorCustomDataSchema import (
    error_message_,
    error_code_,
    error_delegator_custom_data_desc,
    error_delegator_stub_custom_data_example,
)
from argrelay.plugin_delegator.NoopDelegator import NoopDelegator
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_data.AssignedValue import AssignedValue
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    instance_data_,
)
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
    func_id_,
)
from argrelay.schema_config_interp.SearchControlSchema import (
    populate_search_control,
)
from argrelay.schema_response.InvocationInput import InvocationInput

host_container_ipos_ = 1
service_container_ipos_ = 1
access_container_ipos_ = 2


def set_default_to(arg_type, arg_val, envelope_container):
    if arg_type in envelope_container.search_control.types_to_keys_dict:
        if arg_type not in envelope_container.assigned_types_to_values:
            if arg_type in envelope_container.remaining_types_to_values:
                if arg_val in envelope_container.remaining_types_to_values[arg_type]:
                    del envelope_container.remaining_types_to_values[arg_type]
                    envelope_container.assigned_types_to_values[arg_type] = AssignedValue(
                        arg_val,
                        ArgSource.DefaultValue,
                    )
        else:
            # Setting `ArgSource.DefaultValue`: it cannot override any, right? No point to handle assigned case:
            pass


def redirect_to_no_func_error(
    interp_ctx,
    server_config,
):
    return redirect_to_error(
        interp_ctx,
        server_config,
        "ERROR: objects cannot be searched until function is fully qualified",
        1,
    )


def redirect_to_error(
    interp_ctx,
    server_config,
    error_message,
    error_code,
):
    # Redirect to `ErrorDelegator`:
    # TODO: Do not hardcode plugin id (instance of `ErrorDelegator`):
    delegator_plugin_instance_id = f"{ErrorDelegator.__name__}.default"
    custom_plugin_data = {
        error_message_: error_message,
        error_code_: error_code,
    }
    error_delegator_custom_data_desc.validate_dict(custom_plugin_data)
    invocation_input = InvocationInput(
        arg_values = interp_ctx.comp_suggestions,
        all_tokens = interp_ctx.parsed_ctx.all_tokens,
        consumed_tokens = interp_ctx.consumed_tokens,
        envelope_containers = interp_ctx.envelope_containers,
        tan_token_ipos = interp_ctx.parsed_ctx.tan_token_ipos,
        tan_token_l_part = interp_ctx.parsed_ctx.tan_token_l_part,
        delegator_plugin_entry = server_config.plugin_instance_entries[delegator_plugin_instance_id],
        custom_plugin_data = custom_plugin_data,
    )
    return invocation_input


class ServiceDelegator(AbstractDelegator):

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:

        class_to_collection_map: dict = self.server_config.class_to_collection_map

        cluster_search_control = populate_search_control(
            class_to_collection_map,
            ServiceEnvelopeClass.ClassCluster.name,
            [
                {"code": ServiceArgType.CodeMaturity.name},
                {"stage": ServiceArgType.FlowStage.name},
                {"region": ServiceArgType.GeoRegion.name},
                {"cluster": ServiceArgType.ClusterName.name},
            ],
        )

        host_search_control = populate_search_control(
            class_to_collection_map,
            ServiceEnvelopeClass.ClassHost.name,
            [
                # ClassCluster:
                {"code": ServiceArgType.CodeMaturity.name},
                {"stage": ServiceArgType.FlowStage.name},
                {"region": ServiceArgType.GeoRegion.name},
                {"cluster": ServiceArgType.ClusterName.name},
                # ClassHost:
                {"host": ServiceArgType.HostName.name},
                # ---
                {"ip": ServiceArgType.IpAddress.name},
            ],
        )

        service_search_control = populate_search_control(
            class_to_collection_map,
            ServiceEnvelopeClass.ClassService.name,
            [
                # ClassCluster:
                {"code": ServiceArgType.CodeMaturity.name},
                {"stage": ServiceArgType.FlowStage.name},
                {"region": ServiceArgType.GeoRegion.name},
                {"cluster": ServiceArgType.ClusterName.name},
                # ClassService:
                {"group": ServiceArgType.GroupLabel.name},
                {"service": ServiceArgType.ServiceName.name},
                # ClassHost:
                {"host": ServiceArgType.HostName.name},
                # ---
                {"status": ServiceArgType.LiveStatus.name},
                {"dc": ServiceArgType.DataCenter.name},
                {"ip": ServiceArgType.IpAddress.name},
            ],
        )

        access_search_control = populate_search_control(
            class_to_collection_map,
            ServiceArgType.AccessType.name,
            [
                {"access": ServiceArgType.AccessType.name},
            ],
        )

        func_envelopes = [
            {
                instance_data_: {
                    func_id_: goto_host_func_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        host_search_control,
                        access_search_control,
                    ],
                },
                ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedArgType.HelpHint.name: "Go (log in) to remote host",
                ReservedArgType.FuncId.name: goto_host_func_,
            },
            {
                instance_data_: {
                    func_id_: goto_service_func_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        service_search_control,
                        access_search_control,
                    ],
                },
                ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedArgType.HelpHint.name: "Go (log in) to remote host and dir path with specified service",
                ReservedArgType.FuncId.name: goto_service_func_,
            },
            {
                instance_data_: {
                    func_id_: desc_host_func_,
                    # TODO: Do not hardcode plugin id (instance of `NoopDelegator`):
                    delegator_plugin_instance_id_: f"{NoopDelegator.__name__}.default",
                    search_control_list_: [
                        host_search_control,
                    ],
                },
                ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedArgType.HelpHint.name: "Describe remote host",
                ReservedArgType.FuncId.name: desc_host_func_,
            },
            {
                instance_data_: {
                    func_id_: desc_service_func_,
                    # TODO: Do not hardcode plugin id (instance of `NoopDelegator`):
                    delegator_plugin_instance_id_: f"{NoopDelegator.__name__}.default",
                    search_control_list_: [
                        service_search_control,
                    ],
                },
                ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedArgType.HelpHint.name: "Describe service instance",
                ReservedArgType.FuncId.name: desc_service_func_,
            },
            {
                instance_data_: {
                    func_id_: list_host_func_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        host_search_control,
                    ],
                },
                ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedArgType.HelpHint.name: "List remote hosts matching search query",
                ReservedArgType.FuncId.name: list_host_func_,
            },
            {
                instance_data_: {
                    func_id_: list_service_func_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        service_search_control,
                    ],
                },
                ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedArgType.HelpHint.name: "List service instances matching search query",
                ReservedArgType.FuncId.name: list_service_func_,
            },
        ]
        return func_envelopes

    def has_fill_control(
        self,
    ) -> bool:
        return True

    def run_fill_control(
        self,
        interp_ctx: "InterpContext",
    ) -> bool:
        func_id = get_func_id_from_interp_ctx(interp_ctx)
        if func_id in [
            goto_host_func_,
            goto_service_func_,
        ]:
            assert host_container_ipos_ == service_container_ipos_
            object_container_ipos = host_container_ipos_

            # If need to specify `AccessType` `data_envelope`:
            if interp_ctx.curr_container_ipos == interp_ctx.curr_interp.base_container_ipos + access_container_ipos_:
                # Take object found so far:
                data_envelope = interp_ctx.envelope_containers[(
                    interp_ctx.curr_interp.base_container_ipos + object_container_ipos
                )].data_envelopes[0]

                access_container = interp_ctx.envelope_containers[(
                    interp_ctx.curr_interp.base_container_ipos + access_container_ipos_
                )]

                # Select default value to search `AccessType` `data_envelope` based on `CodeMaturity`:
                code_arg_type = ServiceArgType.CodeMaturity.name
                if code_arg_type in data_envelope:
                    code_arg_val = data_envelope[code_arg_type]
                    if code_arg_val == "prod":
                        set_default_to(ServiceArgType.AccessType.name, "ro", access_container)
                    else:
                        set_default_to(ServiceArgType.AccessType.name, "rw", access_container)
                    return True

        elif func_id in [
            list_host_func_,
            list_service_func_,
        ]:
            pass
        else:
            raise RuntimeError

        return False

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        func_id = get_func_id_from_interp_ctx(interp_ctx)

        vararg_container_ipos = host_container_ipos_
        assert vararg_container_ipos == host_container_ipos_ == service_container_ipos_

        if func_id in [
            goto_host_func_,
            goto_service_func_,
        ]:
            # Even if these functions do not support varargs, when `redirect_to_error`, query all:
            vararg_container = interp_ctx.envelope_containers[vararg_container_ipos]
            vararg_container.data_envelopes = (
                local_server
                .get_query_engine()
                .query_data_envelopes_for(vararg_container)
            )

            # Actual implementation is not defined for demo:
            return redirect_to_error(
                interp_ctx,
                local_server.server_config,
                error_delegator_stub_custom_data_example[error_message_],
                error_delegator_stub_custom_data_example[error_code_],
            )
        elif func_id in [
            list_host_func_,
            list_service_func_,
        ]:
            # Verify that func is selected and all what is left to do is to query 0...N objects:
            if interp_ctx.curr_container_ipos >= vararg_container_ipos:
                # Search `data_envelope`-s based on existing args on command line:
                vararg_container = interp_ctx.envelope_containers[vararg_container_ipos]
                vararg_container.data_envelopes = (
                    local_server
                    .get_query_engine()
                    .query_data_envelopes_for(vararg_container)
                )

                # Plugin to invoke on client side:
                delegator_plugin_instance_id = self.plugin_instance_id
                # Package into `InvocationInput` payload object:
                invocation_input = InvocationInput(
                    arg_values = interp_ctx.comp_suggestions,
                    all_tokens = interp_ctx.parsed_ctx.all_tokens,
                    consumed_tokens = interp_ctx.consumed_tokens,
                    envelope_containers = interp_ctx.envelope_containers,
                    tan_token_ipos = interp_ctx.parsed_ctx.tan_token_ipos,
                    tan_token_l_part = interp_ctx.parsed_ctx.tan_token_l_part,
                    delegator_plugin_entry = local_server.server_config.plugin_instance_entries[
                        delegator_plugin_instance_id
                    ],
                    custom_plugin_data = {},
                )
                return invocation_input
            else:
                return redirect_to_no_func_error(
                    interp_ctx,
                    local_server.server_config,
                )
        else:
            # Plugin is given a function name it does not know:
            raise RuntimeError

    @staticmethod
    def invoke_action(invocation_input: InvocationInput):
        """
        Print `data_envelope`-s received from server on client side.
        """

        func_id = get_func_id_from_invocation_input(invocation_input)
        if func_id == list_host_func_:
            for data_envelope in invocation_input.envelope_containers[host_container_ipos_].data_envelopes:
                print(json.dumps(data_envelope))
        elif func_id == list_service_func_:
            for data_envelope in invocation_input.envelope_containers[service_container_ipos_].data_envelopes:
                print(json.dumps(data_envelope))
        else:
            raise RuntimeError



import logging
import os.path
from typing import Union, Callable, Any

import pkg_resources
from flasgger import Swagger
from flask import Flask, request, redirect

from argrelay import relay_server
from argrelay.custom_integ.git_utils import get_git_repo_root_path
from argrelay.misc_helper_common import get_argrelay_dir
from argrelay.plugin_config.AbstractConfigurator import AbstractConfigurator
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.relay_server.route_api import create_blueprint_api
from argrelay.relay_server.route_gui import create_blueprint_gui
from argrelay.schema_config_core_server.ServerConfigSchema import server_config_desc
from argrelay.server_spec.const_int import API_SPEC_PATH, API_DOCS_PATH, ARGRELAY_GUI_PATH

# Set this here (because `require` function may fail in other contexts):
server_version = pkg_resources.require("argrelay")[0].version
server_title = relay_server.__name__


class CustomFlaskApp(Flask):
    """
    This is an API-wrapper exposing `LocalServer` over the network.
    """

    def __init__(
        self,
        import_name: str,
        static_url_path = "/gui_static",
        static_folder = pkg_resources.resource_filename(relay_server.__name__, "gui_static"),
        template_folder = pkg_resources.resource_filename(relay_server.__name__, "gui_templates"),
    ):
        super().__init__(
            import_name = import_name,
            static_folder = static_folder,
            template_folder = template_folder,
            static_url_path = static_url_path,
        )
        self.local_server: LocalServer = LocalServer(server_config_desc.obj_from_default_file())

    def run_with_config(self):
        # Use custom logging at DEBUG level - see: `log_request` and `log_response:
        self.logger.setLevel(logging.DEBUG)
        # Log only at ERROR level by default logger:
        # https://stackoverflow.com/a/18379764/441652
        logging.getLogger("werkzeug").setLevel(logging.ERROR)

        self.run(
            # Contrary to "debug" keyword, if IDE debug is needed, set `debug` to `False` to avoid reloader:
            # https://stackoverflow.com/a/53790400/441652
            # NOTE: Disabled debug to make sure reloader does not restart server (which does not work with Mongo DB):
            debug = False,
            host = self.local_server.server_config.connection_config.server_host_name,
            port = self.local_server.server_config.connection_config.server_port_number,
        )


def create_app() -> CustomFlaskApp:
    """
    The function prevents running the code inside it on import - necessary for pre-test mocking.
    Note that Flask may rely on specific name `create_app` to find this factory:
    https://flask.palletsprojects.com/en/2.2.x/patterns/appfactories/#using-applications
    """

    flask_app = CustomFlaskApp(
        import_name = relay_server.__name__,
    )

    flask_app.local_server.start_local_server()

    swagger_template = {
        "info": {
            "title": server_title,
            "version": server_version,
            "description": f"See <a href=\"{ARGRELAY_GUI_PATH}\" target=\"_blank\">built-in GUI</a>.",
        },
    }

    swagger_config = {
        "headers": [],
        "specs": [
            {
                "endpoint": "relay_server",
                "route": API_SPEC_PATH,
                "rule_filter": lambda rule: True,
                "model_filter": lambda tag: True,
            },
        ],
        "static_url_path": "/flasgger_static",
        "swagger_ui": True,
        "specs_route": API_DOCS_PATH,
    }

    Swagger(
        flask_app,
        template = swagger_template,
        config = swagger_config,
    )

    @flask_app.before_request
    def log_request():
        flask_app.logger.debug(
            "request: %s\n    source: %s\n    headers: %s\n    body: %s",
            request,
            request.remote_addr,
            request.headers.__repr__(),
            request.get_data(),
        )

    @flask_app.after_request
    def log_response(response):
        # Wrap into try/except because of this:
        # https://stackoverflow.com/q/64006669/441652
        # noinspection PyBroadException
        try:
            response_data = str(response.get_data())
        except Exception as e:
            response_data = str(e)
        flask_app.logger.debug(
            "response: %s\n    headers: %s\n    body: %s\n",
            response,
            response.headers.__repr__(),
            response_data,
        )
        return response

    @flask_app.route("/")
    def root_redirect():
        return redirect(ARGRELAY_GUI_PATH, code = 302)

    flask_app.register_blueprint(create_blueprint_api(flask_app.local_server))
    flask_app.register_blueprint(create_blueprint_gui(
        configure_project_title(flask_app.local_server.server_config.server_configurators),
        configure_project_page_url(flask_app.local_server.server_config.server_configurators),
        server_version,
        flask_app.local_server.server_config.gui_banner_config,
        flask_app.local_server.server_start_time,
        # TODO: make AbstractConfiguration expose only these final methods (and any concatenation of strings, validation, should be hidden inside DefaultConfigurator):
        configure_project_git_commit_time(flask_app.local_server.server_config.server_configurators),
        configure_project_git_commit_url(flask_app.local_server.server_config.server_configurators),
        configure_project_git_commit_display_string(flask_app.local_server.server_config.server_configurators),
        configure_project_git_conf_dir_url(flask_app.local_server.server_config.server_configurators),
        configure_project_git_conf_dir_display_string(flask_app.local_server.server_config.server_configurators),
    ))

    return flask_app


def configure_project_title(
    server_configurators,
) -> str:
    return get_config_value_once(
        server_configurators,
        lambda server_configurator: server_configurator.provide_project_title(),
        "[unknown]",
    )


def configure_project_page_url(
    server_configurators,
) -> str:
    return get_config_value_once(
        server_configurators,
        lambda server_configurator: server_configurator.provide_project_page_url(),
        "",
    )


def configure_project_git_commit_time(
    server_configurators,
) -> int:
    return get_config_value_once(
        server_configurators,
        lambda server_configurator: server_configurator.provide_project_git_commit_time(),
        0,
    )


def configure_project_git_commit_url(
    server_configurators,
) -> str:
    project_commit_id_url_prefix: str = get_config_value_once(
        server_configurators,
        lambda server_configurator: server_configurator.provide_project_commit_id_url_prefix(),
        None,
    )
    project_git_commit_id: str = get_config_value_once(
        server_configurators,
        lambda server_configurator: server_configurator.provide_project_git_commit_id(),
        None,
    )
    if (
        project_commit_id_url_prefix is not None
        and
        project_git_commit_id is not None
    ):
        return f"{project_commit_id_url_prefix}{project_git_commit_id}"
    else:
        return ""


def configure_project_git_commit_display_string(
    server_configurators,
) -> str:
    return get_config_value_once(
        server_configurators,
        lambda server_configurator: server_configurator.provide_project_git_commit_display_string(),
        "[unknown]",
    )


def configure_project_git_conf_dir_url(
    server_configurators,
) -> Union[str, None]:
    """
    Full URL is going to be formed only if
    *   `project_current_config_path` points to the same Get repo root as `argrelay_dir`
    *   `project_current_config_path` points under `argrelay_dir`
    """
    project_git_files_by_commit_id_url_prefix = get_config_value_once(
        server_configurators,
        lambda server_configurator: server_configurator.provide_project_git_files_by_commit_id_url_prefix(),
        None,
    )
    project_git_commit_id: str = get_config_value_once(
        server_configurators,
        lambda server_configurator: server_configurator.provide_project_git_commit_id(),
        None,
    )
    project_git_repo_relative_argrelay_dir: str = get_config_value_once(
        server_configurators,
        lambda server_configurator: server_configurator.provide_project_git_repo_relative_argrelay_dir(),
        None,
    )
    project_current_config_path = get_config_value_once(
        server_configurators,
        lambda server_configurator: server_configurator.provide_project_current_config_path(),
        None,
    )

    # Validation:

    if (
        project_git_files_by_commit_id_url_prefix is None
        or
        project_git_commit_id is None
        or
        project_git_repo_relative_argrelay_dir is None
        or
        project_current_config_path is None
    ):
        return ""

    argrelay_dir_abs_path = os.path.realpath(os.path.abspath(get_argrelay_dir()))
    project_current_config_abs_path = os.path.realpath(os.path.abspath(os.path.join(
        argrelay_dir_abs_path,
        project_current_config_path
    )))

    # This may happen even if paths were joined
    # if `project_current_config_path` is a symlink pointing outside `argrelay_dir_abs_path`:
    if not project_current_config_abs_path.startswith(argrelay_dir_abs_path):
        return ""

    # This may happen if `@/conf` is a sub-dir, and it is another git repo:
    if (
        os.path.realpath(get_git_repo_root_path(argrelay_dir_abs_path))
        !=
        os.path.realpath(get_git_repo_root_path(project_current_config_abs_path))
    ):
        return ""

    # Compose URL:
    project_current_config_path = os.path.relpath(project_current_config_abs_path, argrelay_dir_abs_path)
    return f"{project_git_files_by_commit_id_url_prefix}{project_git_commit_id}{project_git_repo_relative_argrelay_dir}{project_current_config_path}"


def configure_project_git_conf_dir_display_string(
    server_configurators,
) -> str:
    project_current_config_path = get_config_value_once(
        server_configurators,
        lambda server_configurator: server_configurator.provide_project_current_config_path(),
        None,
    )
    if project_current_config_path is not None:
        return project_current_config_path
    else:
        return "[unknown]"


def get_config_value_once(
    server_configurators,
    value_getter: Callable[[AbstractConfigurator], Any],
    default_value: Any,
) -> Any:
    config_value: Union[Any, None] = None
    server_configurator: AbstractConfigurator
    for server_configurator in server_configurators.values():
        if config_value is None:
            config_value = value_getter(server_configurator)
        else:
            if value_getter(server_configurator) is not None:
                # Only one `PluginType.ConfiguratorPlugin` providing
                # same type of value is supported to avoid confusion:
                raise RuntimeError

    if config_value is None:
        config_value = default_value
    return config_value



from enum import auto, Enum


class DistinctValuesQuery(Enum):
    """
    Selects implementation for FS_02_25_41_81 `query_enum_items_func`.

    See `test_QueryEngine_perf.py` for perf test used to produce these tables:

    Numbers for "use_single_collection: True" are more evident as `object_count` directly affects the query.

    Numbers for "use_single_collection: False" are less evident as `object_count` is distributed across collections.

    *   `mongomock`:

        ```
        ================================
        use_mongomock: True
        use_single_collection: True

                       object_multiplier         3         4         5         6         7         8         9
                            object_count       243      1024      3125      7776     16807     32768     59049
        --------------------------------
                  original_find_and_loop     0.009     0.030     0.115     0.440     1.632     6.137    33.669
                         native_distinct     0.038     0.122     0.363     0.969     2.183     4.740     7.102
                        native_aggregate     0.044     0.151     0.679     2.899    13.257    49.728   292.153
        ================================
        use_mongomock: True
        use_single_collection: False

                       object_multiplier         3         4         5         6         7         8         9
                            object_count       243      1024      3125      7776     16807     32768     59049
        --------------------------------
                  original_find_and_loop     0.006     0.018     0.113     0.414     1.662     5.831    29.375
                         native_distinct     0.027     0.096     0.327     0.812     1.830     3.578     6.403
                        native_aggregate     0.020     0.066     0.258     1.108     4.923    17.561    95.461
        ```

        *   The fastest on small collections is `original_find_and_loop`.
        *   The fastest on large collections is `native_distinct`.

    *   `pymongo`:

        ```
        ================================
        use_mongomock: False
        use_single_collection: True

                       object_multiplier         3         4         5         6         7         8         9
                            object_count       243      1024      3125      7776     16807     32768     59049
        --------------------------------
                  original_find_and_loop     0.013     0.023     0.043     0.090     0.182     0.428     0.746
                         native_distinct     0.041     0.065     0.117     0.241     0.462     0.796     1.458
                        native_aggregate     0.022     0.027     0.050     0.081     0.125     0.210     0.368
        ================================
        use_mongomock: False
        use_single_collection: False

                       object_multiplier         3         4         5         6         7         8         9
                            object_count       243      1024      3125      7776     16807     32768     59049
        --------------------------------
                  original_find_and_loop     0.015     0.020     0.039     0.087     0.148     0.345     0.633
                         native_distinct     0.044     0.048     0.087     0.173     0.299     0.512     0.903
                        native_aggregate     0.016     0.021     0.033     0.054     0.086     0.159     0.231
        ```

        The fastest is `native_aggregate`.
    """

    original_find_and_loop = auto()
    """
    See `QueryEngine._query_prop_values_original_find_and_loop`.

    It calls  MongoDB `find()` with `query_dict`, then loops to collect distinct values.

    This is the fastest option for small collections with `mongomock`.
    """

    native_distinct = auto()
    """
    See `QueryEngine._query_prop_values_native_distinct`.

    It does multiple calls to MongoDB `distinct()` with `query_dict`
    to collect set of distinct values per field at a time:
    https://stackoverflow.com/a/20655662/441652

    This is the fastest option for large collections with `mongomock`.
    """

    native_aggregate = auto()
    """
    See `QueryEngine._query_prop_values_native_aggregate`.

    It calls MongoDB `aggregate()` with `$match: query_dict`, then `$group` with `$addToSet`:
    *   See `test_MongoClient_distinc_values_search.py`:
        https://stackoverflow.com/a/63595331/441652
    *   Another example:
        https://stackoverflow.com/a/11991854/441652

    This is:
    *   the fastest option with `pymongo`.
    *   the slowest option with `mongomock`.
    """



from marshmallow import Schema, fields, RAISE, post_load, pre_dump

from argrelay.enum_desc.CompScope import CompScope
from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.misc_helper_common import ensure_value_is_enum
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.server_spec.CallContext import CallContext

server_action_ = "server_action"
command_line_ = "command_line"
cursor_cpos_ = "cursor_cpos"
comp_scope_ = "comp_scope"
is_debug_enabled_ = "is_debug_enabled"

_sample_command_line = "some_command goto service "

_call_context_example = {
    server_action_: ServerAction.DescribeLineArgs.name,
    command_line_: _sample_command_line,
    cursor_cpos_: len(_sample_command_line),
    comp_scope_: CompScope.ScopeInitial.name,
    is_debug_enabled_: False,
}


class CallContextSchema(Schema):
    class Meta:
        unknown = RAISE
        ordered = True

    server_action = fields.Enum(
        ServerAction,
        required = True,
        metadata = {
            "description": (
                "Action to perform - see " + ServerAction.__name__ + " enum"
            ),
            "example": _call_context_example[server_action_],
        },
    )
    command_line = fields.String(
        required = True,
        metadata = {
            "example": _call_context_example[command_line_],
        },
    )
    cursor_cpos = fields.Integer(
        required = True,
        metadata = {
            "description": "Cursor position within command line (0 = before the first char)",
            "example": _call_context_example[cursor_cpos_],
        },
    )
    comp_scope = fields.Enum(
        CompScope,
        required = True,
        metadata = {
            "description": "Name for a completion scope - see " + CompScope.__name__ + " enum",
            "example": _call_context_example[comp_scope_],
        },
    )
    is_debug_enabled = fields.Boolean(
        required = True,
        metadata = {
            "description": "Enable extra debug output",
            "example": _call_context_example[is_debug_enabled_],
        },
    )

    @pre_dump
    def make_dict(
        self,
        input_object: CallContext,
        **kwargs,
    ):
        return {
            server_action_: input_object.server_action,
            command_line_: input_object.command_line,
            cursor_cpos_: input_object.cursor_cpos,
            comp_scope_: ensure_value_is_enum(input_object.comp_scope, CompScope),
            is_debug_enabled_: input_object.is_debug_enabled,
        }

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        return CallContext(
            server_action = ensure_value_is_enum(input_dict[server_action_], ServerAction),
            command_line = input_dict[command_line_],
            cursor_cpos = input_dict[cursor_cpos_],
            # TODO: Is calling `ensure_value_is_enum` even required?
            comp_scope = ensure_value_is_enum(input_dict[comp_scope_], CompScope),
            is_debug_enabled = input_dict[is_debug_enabled_],
        )


call_context_desc = TypeDesc(
    dict_schema = CallContextSchema(),
    ref_name = CallContextSchema.__name__,
    dict_example = _call_context_example,
    default_file_path = "",
)



from __future__ import annotations

from marshmallow import Schema, RAISE, fields

from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    data_envelope_desc,
    sample_field_type_A_,
    sample_field_type_B_,
    sample_field_type_C_,
)
from argrelay.schema_response.EnvelopeContainerSchema import data_envelopes_

collection_name_to_index_fields_map_ = "collection_name_to_index_fields_map"

class ConfigOnlyLoaderConfigSchema(Schema):
    """
    Part of FS_49_96_50_77 config_only_loader implementation.
    """

    class Meta:
        unknown = RAISE
        strict = True

    collection_name_to_index_fields_map = fields.Dict(
        keys = fields.String(),
        values = fields.List(
            fields.String(),
            required = True,
        ),
        required = True,
    )

    data_envelopes = fields.List(
        fields.Nested(data_envelope_desc.dict_schema),
        required = True,
    )


config_only_loader_config_desc = TypeDesc(
    dict_schema = ConfigOnlyLoaderConfigSchema(),
    ref_name = ConfigOnlyLoaderConfigSchema.__name__,
    dict_example = {
        collection_name_to_index_fields_map_: {
            ReservedEnvelopeClass.ClassFunction.name: [
                sample_field_type_A_,
                sample_field_type_B_,
                sample_field_type_C_,
            ],
        },
        data_envelopes_: [
            data_envelope_desc.dict_example,
        ],
    },
    default_file_path = "",
)



import os
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any

import yaml
from marshmallow import Schema, ValidationError
from yaml import SafeLoader

from argrelay.misc_helper_common import get_config_path, get_argrelay_dir


def schemaless_dict_from_yaml_file(
    file_path: str,
) -> dict:
    """
    Avoid using this method (directly) as it loads YAML data without `Schema`.

    See `TypeDesc.dict_from_yaml_file` instead.
    """
    with open(file_path) as yaml_file:
        return yaml.safe_load(yaml_file)


def construct_include(
    yaml_loader: SafeLoader,
    yaml_node: yaml.Node,
) -> Any:
    """
    Method which allows `!include`-ing files in YAML.

    See:
    https://stackoverflow.com/a/9577670/441652
    """

    file_name = os.path.abspath(os.path.join(
        # All paths should be relative to `@/` = `argrelay_dir`:
        get_argrelay_dir(),
        yaml_loader.construct_scalar(yaml_node),
    ))

    with open(file_name, "r") as file_stream:
        return yaml.load(file_stream, SafeLoader)


yaml.add_constructor("!include", construct_include, SafeLoader)


@dataclass(frozen = True)
class TypeDesc:
    """
    Type descriptor: schema, example, etc.
    """

    dict_schema: Schema = field()
    ref_name: str = field()
    dict_example: dict = field()
    default_file_path: str = field()

    def get_adjusted_file_path(
        self,
    ) -> str:
        return get_config_path(self.default_file_path)

    def obj_from_default_file(
        self,
    ) -> object:
        return self.obj_from_yaml_file(self.get_adjusted_file_path())

    def dict_from_default_file(
        self,
    ) -> dict:
        return self.dict_from_yaml_file(self.get_adjusted_file_path())

    def obj_from_yaml_file(
        self,
        file_path: str,
    ) -> object:
        return self.obj_from_input_dict(schemaless_dict_from_yaml_file(file_path))

    def dict_from_yaml_file(
        self,
        file_path: str,
    ) -> dict:
        return self.dict_from_input_dict(schemaless_dict_from_yaml_file(file_path))

    def obj_from_yaml_str(
        self,
        yaml_str: str,
    ) -> object:
        return self.obj_from_input_dict(yaml.safe_load(yaml_str))

    def dict_from_yaml_str(
        self,
        yaml_str: str,
    ) -> object:
        return self.dict_from_input_dict(yaml.safe_load(yaml_str))

    def obj_from_input_dict(
        self,
        input_dict: dict,
    ) -> object:
        # Make a `deepcopy` as `marshmallow` reuses instances specified in `load_default`:
        return deepcopy(self.dict_schema.load(input_dict))

    def dict_from_input_dict(
        self,
        input_dict: dict,
    ) -> dict:
        """
        This `dict` to `dict` method is used to make `Schema` populate defaults:
        """
        return self.dict_from_input_obj(
            # Load `obj` first to populate all defaults based on schema definition:
            self.obj_from_input_dict(input_dict)
        )

    def dict_from_input_obj(
        self,
        input_obj: object,
    ) -> dict:
        # Make a `deepcopy` just in case (if `marshmallow` reuses some of the instances):
        return deepcopy(self.dict_schema.dump(input_obj))

    def validate_dict(
        self,
        input_dict: dict,
    ) -> None:
        validation_errors = self.dict_schema.validate(input_dict)
        if validation_errors:
            raise ValidationError(validation_errors)



from __future__ import annotations

from copy import deepcopy

import mongomock
from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.database import Database

from argrelay.misc_helper_common import eprint
from argrelay.mongo_data.MongoConfig import MongoConfig
from argrelay.runtime_data.EnvelopeCollection import EnvelopeCollection
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    envelope_id_,
    data_envelope_desc,
    mongo_id_,
)


def get_mongo_client(mongo_config: MongoConfig):
    if mongo_config.use_mongomock:
        return mongomock.MongoClient()
    else:
        return MongoClient(mongo_config.mongo_client.client_connection_string)


def store_envelopes(
    mongo_db: Database,
    static_data: StaticData,
):
    # Calculate total:
    total_envelope_n: int = 0
    for mongo_collection in static_data.envelope_collections:
        envelope_collection: EnvelopeCollection = static_data.envelope_collections[
            mongo_collection
        ]
        total_envelope_n += len(envelope_collection.data_envelopes)

    # Index all:
    curr_envelope_i: int = 0
    for mongo_collection in static_data.envelope_collections:
        envelope_collection: EnvelopeCollection = static_data.envelope_collections[
            mongo_collection
        ]
        curr_envelope_i += store_envelope_collection(
            mongo_db,
            mongo_collection,
            envelope_collection,
            curr_envelope_i,
            total_envelope_n,
        )

    assert curr_envelope_i == total_envelope_n


def store_envelope_collection(
    mongo_db: Database,
    mongo_collection: str,
    envelope_collection: EnvelopeCollection,
    curr_envelope_i: int,
    total_envelope_n: int,
) -> int:
    col_proxy: Collection = mongo_db[mongo_collection]
    col_proxy.delete_many({})
    col_proxy.drop_indexes()

    base_curr_envelope_i: int = curr_envelope_i
    envelope_per_col_i: int = 0
    log_index_progress(mongo_collection, envelope_per_col_i, curr_envelope_i, total_envelope_n)
    for data_envelope in envelope_collection.data_envelopes:
        if curr_envelope_i % 1_000 == 0:
            log_index_progress(mongo_collection, envelope_per_col_i, curr_envelope_i, total_envelope_n)
        curr_envelope_i += 1
        envelope_per_col_i += 1
        envelope_to_store = deepcopy(data_envelope)

        try:
            data_envelope_desc.validate_dict(envelope_to_store)
        except:
            print(f"envelope_to_store: {envelope_to_store}")

        if envelope_id_ in envelope_to_store:
            envelope_to_store[mongo_id_] = data_envelope[envelope_id_]

        col_proxy.insert_one(envelope_to_store)

    log_index_progress(mongo_collection, envelope_per_col_i, curr_envelope_i, total_envelope_n)

    assert envelope_per_col_i == curr_envelope_i - base_curr_envelope_i
    return envelope_per_col_i


def log_index_progress(
    mongo_collection: str,
    envelope_per_col_i: int,
    curr_envelope_i: int,
    total_envelope_n: int,
):
    try:
        assert envelope_per_col_i <= curr_envelope_i <= total_envelope_n
    finally:
        eprint(
            f"collection: {mongo_collection}: indexed envelopes: {envelope_per_col_i}/{curr_envelope_i}/{total_envelope_n}..."
        )


def create_index(
    mongo_db: Database,
    collection_name: str,
    index_fields: list[str],
):
    col_proxy: Collection = mongo_db[collection_name]

    for index_field in index_fields:
        col_proxy.create_index(index_field)



import os

import setuptools

tests_require = [
    "tox",
    "responses",
    "mongomock",
    "pandas",
    "icecream",
    "jsonpath-ng",
]

# To install these extra dev dependencies:
# pip install --editable .[tests]
extras_require = {
    "tests": tests_require,
}


def list_dir(
    # Path relative to the `setup.py` (relative to the repo root `@/`):
    top_dir_path,
):
    file_paths = []
    for (parent_dir_path, child_dir_names, child_file_names) in os.walk(top_dir_path):
        for child_file_name in child_file_names:
            file_paths.append(os.path.join(
                parent_dir_path,
                child_file_name,
            ))
    # All paths start with `top_dir_path`:
    return file_paths


setuptools.setup(
    name = "argrelay",
    # See `docs/dev_notes/version_format.md`:
    version = "0.6.7",
    author = "uvsmtid",
    author_email = "uvsmtid@gmail.com",
    description = "Tab-completion & data search server = total recall for Bash shell",
    long_description = """
See: https://github.com/argrelay/argrelay
    """,
    long_description_content_type = "text/markdown",
    keywords = "argparse, argcomplete, bash, complete",
    url = "https://github.com/argrelay/argrelay",
    project_urls = {
        "Bug Tracker": "https://github.com/argrelay/argrelay/issues",
    },
    classifiers = [
        "Development Status :: 3 - Alpha",

        "Environment :: Console",
        "Framework :: Flask",
        "Intended Audience :: Information Technology",

        "Topic :: Terminals",
        "Topic :: System :: Shells",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces",
        "Topic :: Text Processing :: Indexing",

        "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",

        "Programming Language :: Python :: 3",

        "License :: OSI Approved :: Apache Software License",

        "Operating System :: POSIX :: Linux",
    ],
    # See sample layout:
    # https://docs.python.org/3/distutils/setupscript.html#installing-package-data
    packages = (
        setuptools.find_packages(
            where = "src",
        ) + [
            "argrelay_docs",
        ]
    ),
    # See:
    # https://docs.python.org/3/distutils/setupscript.html#listing-whole-packages
    # Instead of specifying directory of the specific package:
    # "argrelay": "src/argrelay",
    # Specify directory of "root" package:
    # "": "src",
    # Apparently, this makes `argrelay.egg-info` dir appear in `src` on editable mode rather than in root `.`.
    package_dir = {
        "": "./src/",
        "argrelay_docs": "./",
    },
    package_data = {
        "argrelay": [

            # config files:
            "sample_conf/argrelay.server.yaml",
            "sample_conf/argrelay.client.json",

            # GUI client:
            "relay_server/gui_static/argrelay_client.js",
            "relay_server/gui_static/argrelay_favicon.ico",
            "relay_server/gui_static/argrelay_style.css",
            "relay_server/gui_static/external_link.svg",
            "relay_server/gui_templates/argrelay_main.html",

            # other resource files:
            "custom_integ_res/argrelay_common_lib.bash",
            "custom_integ_res/argrelay_rc.bash",
            "custom_integ_res/bootstrap_dev_env.bash",
            "custom_integ_res/dev_shell.bash",
            "custom_integ_res/init_shell_env.bash",
            "custom_integ_res/upgrade_all_packages.bash",

        ],
        "argrelay_docs": list_dir("./docs/") + [
            "readme.md",
        ],
    },
    include_package_data = True,
    python_requires = ">=3.7",
    install_requires = [
        "Flask",
        "Werkzeug",
        "PyYaml",
        "jsonschema",
        "flasgger",
        "marshmallow",
        "apispec",
        "pymongo",
        "GitPython",
        # Use `mongomock` as replacement for Mongo DB in simple prod cases:
        "mongomock",
        "requests",
        "cachetools",
    ],
    tests_require = tests_require,
    extras_require = extras_require,
)



import os
import sys
import typing
from typing import Union


def split_process() -> (
    bool,
    int,
    Union[None, typing.TextIO],
):
    """
    Part of FS_14_59_14_06 pending requests implementation.

    Returns tuple:
    *   bool: whether it is a parent (True) or a child (False)
    *   int: child PID for parent, None for child
    *   int: open file for the parent-side end of the read-only pipe with stdout of the child process, 0 for child

    The func splits parent and child processes setting up a communication pipe between the two.

    The pipe works in the single direction from child to parent:
    *   child connects its stdout to the pipe (instead of a terminal)
    *   parent uses the pipe to read results from the child, but does it only after child completes

    This way, while waiting for the child to complete, the parent can print anything to the terminal
    without interleaving with the child stdout.

    Both child and parent still keep their stderr connected to the terminal (this output may interleave).

    NOTE: Technically, using the extra pipe between parent and child is not required
          for `ServerAction.ProposeArgValues` = Tab because Bash already captures
          stdout in that case until client completes.
          But implementation is uniform here for simplicity.
    """
    # Create pipe for the child to write results as its stdout:
    r_child_stdout_fd, w_child_stdout_fd = os.pipe()
    child_pid: int = os.fork()

    if child_pid > 0:
        os.close(w_child_stdout_fd)
        # Parent reads from the pipe to the child (later, when child has completed):
        child_stdout = os.fdopen(r_child_stdout_fd)
        return (
            True,
            child_pid,
            child_stdout,
        )
    elif child_pid == 0:
        os.close(r_child_stdout_fd)
        # Child writes to the pipe (instead of the terminal):
        sys.stdout = os.fdopen(w_child_stdout_fd, "w")
        return (
            False,
            0,
            None,
        )
    else:
        raise RuntimeError("could not fork child process")



"""
"""
import os
from contextlib import contextmanager
from inspect import getframeinfo, currentframe
from pathlib import PurePath

test_data_ = "test_data"

# Enable/disable breakpoints programmatically:
__is_breakpoint_enabled = False


def enable_breakpoint():
    global __is_breakpoint_enabled
    __is_breakpoint_enabled = True


def disable_breakpoint():
    global __is_breakpoint_enabled
    __is_breakpoint_enabled = False


def is_breakpoint_enabled() -> bool:
    return __is_breakpoint_enabled


def line_no() -> int:
    """
    Get source line in the caller frame
    """
    return getframeinfo(currentframe().f_back).lineno


def line_no_from_ctor() -> int:
    """
    Same as `line_no` but used inside `TestCase` ctor (behind one more stack frame).
    """
    return getframeinfo(currentframe().f_back.f_back).lineno


def parse_line_and_cpos(test_line: str) -> (str, int):
    """
    Translate test line with pipe as cursor place into command line string and cursor char position (cpos)
    """
    assert test_line.count('|') == 1
    cursor_cpos = test_line.index('|')
    command_line = test_line.replace('|', "")
    return command_line, cursor_cpos


@contextmanager
def change_to_known_repo_path(path_from_repo_root = "./tests"):
    """
    This function changes current dir to known path within repo root.

    This allows any other code accessing files by relative paths rely on the stable path within the repo.
    """

    old_pwd = os.getcwd()
    try:
        # When IDE runs, CWD is somewhere under "tests".
        # Keep climbing up until directory becomes "tests":
        if "tests" in PurePath(os.getcwd()).parts:
            while os.path.basename(os.getcwd()) != "tests":
                os.chdir("..")

        # When `tox` runs, CWD = [repo root]:
        if os.path.basename(os.getcwd()) != "tests":
            os.chdir("tests")

        # Now, when same baseline is set regardless `tox` or IDE, change to repo root:
        os.chdir("..")

        # Desired path:
        os.chdir(path_from_repo_root)
        yield
    finally:
        os.chdir(old_pwd)



import os
import signal
import time
import typing
from random import randrange

from argrelay.client_spec.ShellContext import ShellContext
from argrelay.enum_desc.TermColor import TermColor
from argrelay.misc_helper_common import eprint

# Use regular Tab indent size:
spinner_length: int = 4

# Rotate spinner state every:
spinner_frame_sec: float = 0.1

# Results of multiple stream reads:
child_stdout_chunks: list = []

child_stderr_chunk_max_size: int = 100


def signal_handler(signal_number, signal_frame):
    if signal_number == signal.SIGCHLD:
        # The child exited:
        pass


def spin_wait_for_child(
    child_pid: int,
    child_stdout: typing.TextIO,
    shell_ctx: ShellContext,
    spinless_sleep_sec: float,
):
    """
    Display spinner while child request is running.
    """

    signal.signal(signal.SIGCHLD, signal_handler)
    # Prevent spinner on instant replies - have an initial sleep first
    # before drawing spinner for the first time
    # (this sleep will exit prematurely if the child exits):
    time.sleep(spinless_sleep_sec)
    if not is_child_running(child_pid):
        return

    # We waited enough - start spinner:
    pending_cursor = generate_pending_cursor()
    is_child_running_var: bool = is_child_running(child_pid)
    while (
        is_child_running_var
        or
        # Apparently, child block on pipe write (and never exits) if parent does not read it:
        read_next_child_stdout_chunk(child_stdout)
    ):
        # Print spinner state chars:
        eprint(next(pending_cursor), end = "", flush = True)
        if is_child_running_var:
            # Sleep only while child is running:
            time.sleep(spinner_frame_sec)
        # Step backward:
        eprint("\b" * spinner_length, end = "")
        if is_child_running_var:
            # Child exit status can be collected only once:
            is_child_running_var = is_child_running(child_pid)

    # Overwrite spinner state chars by expected content from the prev command line
    # (padded with spaces, if beyond the end of the line):
    first_cpos = shell_ctx.cursor_cpos
    last_cpos = shell_ctx.cursor_cpos + spinner_length
    line_portion = shell_ctx.command_line[first_cpos:last_cpos]
    pad_length = spinner_length - len(line_portion)
    eprint(line_portion + " " * pad_length, end = "")
    # Step backward:
    eprint("\b" * spinner_length, end = "", flush = True)


def is_child_running(
    child_pid: int,
):
    (
        child_pid,
        child_status,
    ) = os.waitpid(child_pid, os.WNOHANG)
    if child_pid == 0 and child_status == 0:
        return True
    else:
        return False


def generate_pending_cursor():
    cursor_states = [
        f"{TermColor.spinner_color.value}/|\\|{TermColor.reset_style.value}",
        f"{TermColor.spinner_color.value}|/|\\{TermColor.reset_style.value}",
        f"{TermColor.spinner_color.value}\\|/|{TermColor.reset_style.value}",
        f"{TermColor.spinner_color.value}|\\|/{TermColor.reset_style.value}",
    ]
    # Use random start state:
    random_shift = randrange(len(cursor_states))
    shifted_states = cursor_states[random_shift:] + cursor_states[:random_shift]
    while True:
        for cursor_state in shifted_states:
            yield cursor_state


def read_next_child_stdout_chunk(
    child_stdout: typing.TextIO,
) -> bool:
    child_stdout_chunk = child_stdout.read(child_stderr_chunk_max_size)
    if child_stdout_chunk:
        child_stdout_chunks.append(child_stdout_chunk)
        return True
    else:
        return False


def print_child_stdout_chunks():
    for child_stdout_chunk in child_stdout_chunks:
        print(child_stdout_chunk, flush = True, end = "")



from argrelay.enum_desc.PluginType import PluginType
from argrelay.runtime_context.AbstractPlugin import AbstractPlugin
from argrelay.runtime_data.StaticData import StaticData


class AbstractLoader(AbstractPlugin):

    def get_plugin_type(
        self,
    ) -> PluginType:
        return PluginType.LoaderPlugin

    def update_static_data(
        self,
        static_data: StaticData,
    ) -> StaticData:
        pass



from marshmallow import fields, validates_schema, INCLUDE

from argrelay.enum_desc.ReservedArgType import ReservedArgType
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    DataEnvelopeSchema,
    envelope_id_,
    instance_data_,
    envelope_payload_,
    sample_field_type_A_,
    sample_field_type_B_,
    sample_field_type_C_,
)
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import function_envelope_instance_data_desc


class FuncEnvelopeSchema(DataEnvelopeSchema):
    class Meta:
        # Same as in base `DataEnvelopeSchema` - all unknown fields used as search props:
        unknown = INCLUDE
        strict = True

    envelope_payload = fields.Dict(
        required = False,
    )

    @validates_schema
    def validate_known(
        self,
        input_dict: dict,
        **kwargs,
    ):
        assert input_dict.get(ReservedArgType.EnvelopeClass.name, None) == ReservedEnvelopeClass.ClassFunction.name
        function_envelope_instance_data_desc.validate_dict(input_dict[instance_data_])


func_id_some_func_ = "func_id_some_func"

func_envelope_desc = TypeDesc(
    dict_schema = DataEnvelopeSchema(),
    ref_name = DataEnvelopeSchema.__name__,
    dict_example = {
        # TODO: Do we need to duplicate `envelope_id_` and `ReservedArgType.FuncId.name`?
        envelope_id_: func_id_some_func_,
        instance_data_: function_envelope_instance_data_desc.dict_example,
        envelope_payload_: {},
        sample_field_type_A_: "A_value_1",
        sample_field_type_B_: "B_value_1",
        sample_field_type_C_: "C_value_1",
        ReservedArgType.EnvelopeClass.name: ReservedEnvelopeClass.ClassFunction.name,
        ReservedArgType.HelpHint.name: f"Some help hint",
        ReservedArgType.FuncId.name: func_id_some_func_,
    },
    default_file_path = "",
)



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass(frozen = True)
class ServerPluginControl:
    """
    Configures (controls) usage of plugins by the server.

    See also `ServerPluginControlSchema.py`.
    """

    first_interp_factory_id: str = field()



from enum import Enum, auto


# TODO: TODO_66_66_75_78: Rename to `ReservedPropType` (or `ReservedPropName`?) - see term dictionary.
class ReservedArgType(Enum):
    """
    Some of the `data_envelope` props used to implement various `argrelay` features.
    """

    EnvelopeClass = auto()

    FuncId = auto()
    """
    Used by FS_26_43_73_72 func tree to distinguish leaves (func ids).
    It is normally the same with `FunctionEnvelopeInstanceDataSchema.func_id_`.
    """

    ArgType = auto()

    ArgValue = auto()

    HelpHint = auto()



from __future__ import annotations

from marshmallow import Schema, RAISE, fields, validates_schema, ValidationError, post_load

from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_context.SearchControl import SearchControl

collection_name_ = "collection_name"

envelope_class_ = "envelope_class"
"""
Specifies `ReservedArgType.EnvelopeClass` to search.
"""

keys_to_types_list_ = "keys_to_types_list"
"""
List of keys to use for named args during interpretation and arg types to use in query.
"""


class SearchControlSchema(Schema):
    """
    Implements FS_31_70_49_15 # search_control

    Schema for search_control which specifies what arg types will be used in search of the `data_envelope`
    (and how they are mapped to named arg keys).
    """

    class Meta:
        unknown = RAISE
        strict = True

    collection_name = fields.String(
        required = True,
    )

    envelope_class = fields.String(
        required = True,
    )

    # FS_20_88_05_60: named args: maps `arg_name` (in command line) to `prop_name` (in `data_envelope`).
    # Each `list` item must contain singleton `dict` (single key-value pair).
    keys_to_types_list = fields.List(
        fields.Dict(
            # `arg_name`:
            keys = fields.String(),
            # `prop_name`:
            values = fields.String(),
            required = True,
        ),
        required = True,
    )

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        return SearchControl(
            collection_name = input_dict[collection_name_],
            envelope_class = input_dict[envelope_class_],
            keys_to_types_list = input_dict[keys_to_types_list_],
        )

    @validates_schema
    def validate_known(
        self,
        input_dict: dict,
        **kwargs,
    ):
        for key_to_type_entry in input_dict[keys_to_types_list_]:
            # ensure there is only one key per dict:
            if len(key_to_type_entry) != 1:
                raise ValidationError("only one key per dict is allowed")


search_control_desc = TypeDesc(
    dict_schema = SearchControlSchema(),
    ref_name = SearchControlSchema.__name__,
    dict_example = {
        collection_name_: ReservedEnvelopeClass.ClassFunction.name,
        envelope_class_: ReservedEnvelopeClass.ClassFunction.name,
        keys_to_types_list_: [
            {
                "type_a": "TypeA",
            },
            {
                "type_b": "TypeB",
            },
        ],
    },
    default_file_path = "",
)


def populate_search_control(
    class_to_collection_map: dict,
    class_name: str,
    keys_to_types_list: list[dict],
) -> dict:
    # If not overridden, each class name uses collection by its own name:
    class_to_collection_map.setdefault(
        class_name,
        class_name,
    )

    search_control: dict = {
        collection_name_: class_to_collection_map[class_name],
        envelope_class_: class_name,
        keys_to_types_list_: keys_to_types_list,
    }

    return search_control



from __future__ import annotations

import subprocess
from typing import Union

_orig_Popen = subprocess.Popen


class PopenMock:
    """
    This class mocks expectation of `subprocess.Popen` based on config in `expected_args_to_output`.
    If it sees unexpected CLI args (not specified in `expected_args_to_output`), it throws an error.
    Otherwise:
    *   If configured tuple is not `None`, it delegates calls to `ConfiguredPopenMockDelegatee`
        (which returns configured tuple (returncode, stdout, stderr) to the `Popen` caller).
    *   If configured tuple is `None`, it passes calls through to the original `Popen` (see `_orig_Popen`).

    Ideas for inspiration:
    *   https://stackoverflow.com/a/53793739/441652
    *   https://stackoverflow.com/a/25693097/441652
    """

    def __init__(
        self,
        expected_args_to_output: dict[Union[str, tuple[str, ...]], tuple[int, str, str]],
        *args,
        **kwargs,
    ):
        self.expected_args_to_output = expected_args_to_output
        """
        Collection of expected input lists of CLI args (`tuple`-s of `str` as keys to `dict`) mapped to
        the output `tuple` of (exit_code, stdout_str, stderr_str).
        """

    def __call__(
        self,
        cli_args,
        *args,
        **kwargs,
    ):
        if isinstance(cli_args, str):
            # Popen with `shell = True` is called with single string arg (full command line).
            # Nevertheless, the config for expected args should specify it as singleton `tuple` (instead of `str`):
            config_key = cli_args
        elif isinstance(cli_args, list):
            config_key = tuple(cli_args)
        else:
            raise ValueError(f"unknown CLI args type: {type(cli_args)}")

        if config_key in self.expected_args_to_output:
            p_output_config_tuple = self.expected_args_to_output[config_key]
            if p_output_config_tuple is None:
                return _orig_Popen(cli_args, *args, **kwargs)
            else:
                # The call to `Popen` with CLI args matching one of the expected ones:
                # all these fields will (potentially) be inspected through other calls to `Popen` - configure them:
                return ConfiguredPopenMockDelegatee(
                    cli_args,
                    *p_output_config_tuple,
                )
        else:
            raise ValueError(f"unexpected CLI args: `{cli_args}` expected CLI args: `{self.expected_args_to_output.keys()}`")


class ConfiguredPopenMockDelegatee:

    def __init__(
        self,
        args: Union[str, list[str]],
        returncode: int,
        stdout: str,
        stderr: str,
    ):
        self.args = args
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr

    def __enter__(
        self,
        *args,
        **kwargs,
    ):
        return self

    def __exit__(
        self,
        *args,
        **kwargs,
    ):
        pass

    def wait(
        self,
        *args,
        **kwargs,
    ):
        return self.returncode

    def kill(
        self,
        *args,
        **kwargs,
    ):
        pass

    def communicate(
        self,
        *args,
        **kwargs,
    ):
        return (
            self.stdout,
            self.stderr,
        )

    def poll(
        self,
        *args,
        **kwargs,
    ):
        return self.returncode
